<?php

namespace Drupal\Tests\jsonapi_schema\Kernel;

use Drupal\Component\Serialization\Json;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\KernelTests\KernelTestBase;
use Drupal\node\Entity\NodeType;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;

/**
 * Class to test JSON:API Schema.
 *
 * @group jsonapi_schema
 */
class JsonApiSchemaTest extends KernelTestBase {

  /**
   * Define expected resource information for tests.
   */
  protected array $expectedResources = [
    'Fields' => [
      'entity_type' => 'field_config',
      'bundle' => 'field_config',
      'title' => 'Field',
      'relationships' => [],
    ],
    'Field storages' => [
      'entity_type' => 'field_storage_config',
      'bundle' => 'field_storage_config',
      'title' => 'Field storage',
      'relationships' => [],
    ],
    'Files' => [
      'entity_type' => 'file',
      'bundle' => 'file',
      'title' => 'File',
      'relationships' => [
        'uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
      ],
    ],
    'Article content items' => [
      'entity_type' => 'node',
      'bundle' => 'article',
      'title' => 'Article content item',
      'relationships' => [
        'node_type' => [
          'title' => 'Content type',
          'entity_type' => 'node_type',
          'bundle' => 'node_type',
          'cardinality' => 1,
        ],
        'revision_uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
        'uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
        'test_entity_reference' => [
          'title' => 'Article content items and Blog content items',
          'entity_type' => 'node',
          'bundle' => ['article', 'blog'],
          'cardinality' => -1,
        ],
        'test_entity_reference2' => [
          'title' => 'Article content item, Blog content item, and Page content item',
          'entity_type' => 'node',
          'bundle' => ['article', 'blog', 'page'],
          'cardinality' => 1,
        ],
      ],
    ],
    'Blog content items' => [
      'entity_type' => 'node',
      'bundle' => 'blog',
      'title' => 'Blog content item',
      'relationships' => [
        'node_type' => [
          'title' => 'Content type',
          'entity_type' => 'node_type',
          'bundle' => 'node_type',
          'cardinality' => 1,
        ],
        'revision_uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
        'uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
      ],
    ],
    'Page content items' => [
      'entity_type' => 'node',
      'bundle' => 'page',
      'title' => 'Page content item',
      'relationships' => [
        'node_type' => [
          'title' => 'Content type',
          'entity_type' => 'node_type',
          'bundle' => 'node_type',
          'cardinality' => 1,
        ],
        'revision_uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
        'uid' => [
          'title' => 'User',
          'entity_type' => 'user',
          'bundle' => 'user',
          'cardinality' => 1,
        ],
      ],
    ],
    'Content types' => [
      'entity_type' => 'node_type',
      'bundle' => 'node_type',
      'title' => 'Content type',
      'relationships' => [],
    ],
    'Menus' => [
      'entity_type' => 'menu',
      'bundle' => 'menu',
      'title' => 'Menu',
      'relationships' => [],
    ],
    'Actions' => [
      'entity_type' => 'action',
      'bundle' => 'action',
      'title' => 'Action',
      'relationships' => [],
    ],
    'Roles' => [
      'entity_type' => 'user_role',
      'bundle' => 'user_role',
      'title' => 'Role',
      'relationships' => [],
    ],
    'Users' => [
      'entity_type' => 'user',
      'bundle' => 'user',
      'title' => 'User',
      'relationships' => [
        'roles' => [
          'title' => 'Roles',
          'entity_type' => 'user_role',
          'bundle' => 'user_role',
          'cardinality' => -1,
        ],
      ],
    ],
    'Base field override entities' => [
      'entity_type' => 'base_field_override',
      'bundle' => 'base_field_override',
      'title' => 'Base field override',
      'relationships' => [],
    ],
    'Date format entities' => [
      'entity_type' => 'date_format',
      'bundle' => 'date_format',
      'title' => 'Date format',
      'relationships' => [],
    ],
    'Entity form display entities' => [
      'entity_type' => 'entity_form_display',
      'bundle' => 'entity_form_display',
      'title' => 'Entity form display',
      'relationships' => [],
    ],
    'View mode entities' => [
      'entity_type' => 'entity_view_mode',
      'bundle' => 'entity_view_mode',
      'title' => 'View mode',
      'relationships' => [],
    ],
    'Form mode entities' => [
      'entity_type' => 'entity_form_mode',
      'bundle' => 'entity_form_mode',
      'title' => 'Form mode',
      'relationships' => [],
    ],
    'Entity view display entities' => [
      'entity_type' => 'entity_view_display',
      'bundle' => 'entity_view_display',
      'title' => 'Entity view display',
      'relationships' => [],
    ],
  ];

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'datetime',
    'field',
    'file',
    'jsonapi',
    'jsonapi_schema',
    'jsonapi_schema_test',
    'node',
    'options',
    'serialization',
    'system',
    'user',
  ];

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();
    $this->installEntitySchema('file');
    $this->installEntitySchema('node');

    // Create a basic page node type.
    NodeType::create([
      'type' => 'page',
      'name' => 'Page',
    ])->save();

    // Create a basic blog node type.
    NodeType::create([
      'type' => 'blog',
      'name' => 'Blog',
    ])->save();

    // Create a basic article node type.
    NodeType::create([
      'type' => 'article',
      'name' => 'Article',
    ])->save();

    // Add an email field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_email',
      'type' => 'email',
      'entity_type' => 'node',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_email',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test email',
      'description' => 'Test email description',
    ]);
    $field->save();

    // Add an entity reference field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_entity_reference',
      'type' => 'entity_reference',
      'entity_type' => 'node',
      'cardinality' => -1,
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_entity_reference',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test entity reference',
      'description' => 'Test entity reference description',
      'required' => TRUE,
      'settings' => [
        'target_type' => 'node',
        'handler_settings' => [
          'target_bundles' => [
            'article',
            'blog',
          ],
        ],
      ],
    ]);
    $field->save();

    // Add a second entity_reference field that allows referencing 3 bundles.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_entity_reference2',
      'type' => 'entity_reference',
      'entity_type' => 'node',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_entity_reference2',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test entity reference 2',
      'description' => 'Test entity reference 2 description',
      'settings' => [
        'target_type' => 'node',
        'handler_settings' => [
          'target_bundles' => [
            'article',
            'blog',
            'page',
          ],
        ],
      ],
    ]);
    $field->save();

    // Add a datetime field with "Date only".
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_date',
      'type' => 'datetime',
      'entity_type' => 'node',
      'settings' => [
        'datetime_type' => 'date',
      ],
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_date',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test date',
      'description' => 'Test date description',
    ]);
    $field->save();

    // Add a datetime field with "Date and time".
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_datetime',
      'type' => 'datetime',
      'entity_type' => 'node',
      'settings' => [
        'datetime_type' => 'datetime',
      ],
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_datetime',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test datetime',
      'description' => 'Test datetime description',
    ]);
    $field->save();

    // Add a decimal field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_decimal',
      'type' => 'decimal',
      'entity_type' => 'node',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_decimal',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test decimal',
      'description' => 'Test decimal description',
    ]);
    $field->save();

    // Add a float field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_float',
      'type' => 'float',
      'entity_type' => 'node',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_float',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test float',
      'description' => 'Test float description',
    ]);
    $field->save();

    // Add a multi-value string field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_string',
      'type' => 'string',
      'entity_type' => 'node',
      'cardinality' => -1,
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_string',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test string',
      'description' => 'Test string description',
    ]);
    $field->save();

    // Add a list_string field with allowed values.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_allowed_values',
      'type' => 'list_string',
      'entity_type' => 'node',
      'settings' => [
        'allowed_values' => [
          'option1' => 'Option 1',
          'option2' => 'Option 2',
          'option3' => 'Option 3',
        ],
      ],
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_allowed_values',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test allowed values',
      'description' => 'Test allowed values description',
    ]);
    $field->save();

    // Add a URI field.
    $field_storage = FieldStorageConfig::create([
      'field_name' => 'test_uri',
      'type' => 'uri',
      'entity_type' => 'node',
    ]);
    $field_storage->save();
    $field = FieldConfig::create([
      'field_name' => 'test_uri',
      'entity_type' => 'node',
      'bundle' => 'article',
      'label' => 'Test URI',
      'description' => 'Test URI description',
    ]);
    $field->save();
  }

  /**
   * Test that /jsonapi/schema is available and has the expected content.
   */
  public function testJsonApiSchemaEntrypoint() {
    $data = $this->apiSchemaRequest('/jsonapi/schema');
    $this->assertNotEmpty($data['allOf']);
    $this->assertEquals('https://jsonapi.org/schema#/definitions/success', $data['allOf'][0]['$ref']);
    $this->assertEquals('object', $data['allOf'][1]['type']);
    $this->assertCount(count($this->expectedResources), $data['allOf'][1]['links']);
    foreach ($data['allOf'][1]['links'] as $resource) {
      $this->assertArrayHasKey($resource['title'], $this->expectedResources);
      $resource_info = $this->expectedResources[$resource['title']];
      $this->assertEquals('{instanceHref}', $resource['href']);
      $this->assertEquals('related', $resource['rel']);
      $this->assertEquals('application/vnd.api+json', $resource['targetMediaType']);
      $this->assertEquals('http://localhost/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/collection/schema', $resource['targetSchema']['$ref']);
      $this->assertEquals('/links/' . $resource_info['entity_type'] . '--' . $resource_info['bundle'] . '/href', $resource['templatePointers']['instanceHref']);
      $this->assertEquals('instanceHref', $resource['templateRequired'][0]);
    }
  }

  /**
   * Test that resource collection schemas are generated.
   */
  public function testJsonApiSchemaCollections() {
    foreach ($this->expectedResources as $title => $resource_info) {
      $data = $this->apiSchemaRequest('/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/collection/schema');
      $this->assertEquals($title, $data['title']);
      $this->assertEquals('https://jsonapi.org/schema', $data['allOf'][0]['$ref']);
      $this->assertEquals('https://jsonapi.org/schema#/definitions/success', $data['allOf'][1]['if']['$ref']);
      $this->assertEquals('object', $data['allOf'][1]['then']['type']);
      $this->assertEquals('#/definitions/data', $data['allOf'][1]['then']['properties']['data']['$ref']);
      $this->assertEquals('data', $data['allOf'][1]['then']['required'][0]);
      $this->assertEquals('array', $data['definitions']['data']['type']);
      $this->assertEquals('http://localhost/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/resource/schema', $data['definitions']['data']['items']['$ref']);
    }
  }

  /**
   * Test that resource schemas are generated.
   *
   * This just checks general data that is shared across all resource types.
   */
  public function testJsonApiSchemaResources() {
    foreach ($this->expectedResources as $resource_info) {
      $data = $this->apiSchemaRequest('/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/resource/schema');
      $this->assertEquals($resource_info['title'], $data['title']);
      $this->assertEquals('object', $data['allOf'][0]['type']);
      $this->assertEquals('#/definitions/type', $data['allOf'][0]['properties']['type']['$ref']);
      $this->assertEquals('#/definitions/attributes', $data['allOf'][0]['properties']['attributes']['$ref']);
      if (!empty($resource_info['relationships'])) {
        $this->assertEquals('#/definitions/relationships', $data['allOf'][0]['properties']['relationships']['$ref']);
      }
      $this->assertEquals('https://jsonapi.org/schema#/definitions/resource', $data['allOf'][1]['$ref']);
      $this->assertEquals($resource_info['entity_type'] . '--' . $resource_info['bundle'], $data['definitions']['type']['const']);

      // Test attributes structure and metadata.
      // Some entity types don't have attributes type or description.
      if (!in_array($resource_info['entity_type'], [
        'base_field_override',
        'entity_form_display',
        'entity_view_mode',
        'entity_form_mode',
        'entity_view_display',
        'field_config',
        'field_storage_config',
      ])) {
        $this->assertEquals('object', $data['definitions']['attributes']['type']);
        $this->assertEquals('Entity attributes', $data['definitions']['attributes']['description']);
      }
      $this->assertNotEmpty($data['definitions']['attributes']['properties']);
      $this->assertFalse($data['definitions']['attributes']['additionalProperties']);

      // Test relationships structure and metadata.
      if (!empty($resource_info['relationships'])) {

        // Test the relationships data type and description.
        $this->assertEquals('Entity relationships', $data['definitions']['relationships']['description']);
        $this->assertEquals('object', $data['definitions']['relationships']['type']);

        // Test that each relationship has the correct metadata.
        // This only checks things that are shared across resource types.
        $this->assertNotEmpty($data['definitions']['relationships']['properties']);
        foreach ($resource_info['relationships'] as $rel => $rel_info) {

          // Test that the relationship is present.
          $this->assertArrayHasKey($rel, $data['definitions']['relationships']['properties']);

          // Test relationships.properties.$name.type.
          $this->assertEquals('object', $data['definitions']['relationships']['properties'][$rel]['type']);

          // The schema structure depends on the relationship cardinality, so we
          // vary the item that we validate accordingly.
          if ($rel_info['cardinality'] === 1) {
            $item = $data['definitions']['relationships']['properties'][$rel]['properties']['data'];
          }
          else {
            $this->assertEquals('array', $data['definitions']['relationships']['properties'][$rel]['properties']['data']['type']);
            $this->assertNotEmpty($data['definitions']['relationships']['properties'][$rel]['properties']['data']['items']);
            $item = $data['definitions']['relationships']['properties'][$rel]['properties']['data']['items'];
          }

          // Test $item.type.
          $this->assertEquals('object', $item['type']);

          // Test $item.required.
          $this->assertEquals([
            'type',
            'id',
          ], $item['required']);

          // Test $item.properties.type.*.
          $this->assertEquals('string', $item['properties']['type']['type']);
          $this->assertEquals('Referenced resource', $item['properties']['type']['title']);

          // Test $item.id.*.
          $this->assertNotEmpty($item['properties']['type']['enum']);
          $this->assertEquals('string', $item['properties']['id']['type']);
          $this->assertEquals('Resource ID', $item['properties']['id']['title']);
          $this->assertEquals('uuid', $item['properties']['id']['format']);
          $this->assertEquals(128, $item['properties']['id']['maxLength']);

          // Test relationships.properties.$name.links.*.
          $this->assertEquals('{instanceHref}', $data['definitions']['relationships']['properties'][$rel]['links'][0]['href']);
          $this->assertEquals('related', $data['definitions']['relationships']['properties'][$rel]['links'][0]['rel']);
          $this->assertEquals('application/vnd.api+json', $data['definitions']['relationships']['properties'][$rel]['links'][0]['targetMediaType']);
          $this->assertEquals('http://localhost/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/resource/relationships/' . $rel . '/related/schema', $data['definitions']['relationships']['properties'][$rel]['links'][0]['targetSchema']['$ref']);
          $this->assertEquals('/links/related/href', $data['definitions']['relationships']['properties'][$rel]['links'][0]['templatePointers']['instanceHref']);
          $this->assertEquals('instanceHref', $data['definitions']['relationships']['properties'][$rel]['links'][0]['templateRequired'][0]);
        }

        // Test that there are no additional properties.
        $this->assertFalse($data['definitions']['relationships']['additionalProperties']);
      }
    }
  }

  /**
   * Test that the node--article resource schema has the expected content.
   *
   * This performs more targeted checks against a specific resource type and its
   * expected field schema.
   */
  public function testJsonApiSchemaNodeResource() {

    // Request resource schema for node--article.
    $data = $this->apiSchemaRequest('/jsonapi/node/article/resource/schema');

    // First we will test to be sure all the expected attributes are present
    // and have the proper type, title, and (optionally) description.
    $expected_attributes = [
      'drupal_internal__nid' => [
        'type' => 'integer',
        'title' => 'ID',
        'cardinality' => 1,
      ],
      'drupal_internal__vid' => [
        'type' => 'integer',
        'title' => 'Revision ID',
        'cardinality' => 1,
      ],
      'langcode' => [
        'type' => 'string',
        'title' => 'Language',
        'cardinality' => 1,
        'allowed_values' => [
          'en' => 'English',
          'und' => 'Not specified',
          'zxx' => 'Not applicable',
        ],
      ],
      'revision_timestamp' => [
        'type' => 'string',
        'title' => 'Revision create time',
        'description' => 'The time that the current revision was created.',
        'cardinality' => 1,
      ],
      'revision_log' => [
        'type' => 'string',
        'title' => 'Revision log message',
        'description' => 'Briefly describe the changes you have made.',
        'cardinality' => 1,
      ],
      'status' => [
        'type' => 'boolean',
        'title' => 'Published',
        'cardinality' => 1,
      ],
      'title' => [
        'type' => 'string',
        'title' => 'Title',
        'cardinality' => 1,
      ],
      'created' => [
        'type' => 'string',
        'title' => 'Authored on',
        'description' => 'The date and time that the content was created.',
        'cardinality' => 1,
      ],
      'changed' => [
        'type' => 'string',
        'title' => 'Changed',
        'description' => 'The time that the node was last edited.',
        'cardinality' => 1,
      ],
      'promote' => [
        'type' => 'boolean',
        'title' => 'Promoted to front page',
        'cardinality' => 1,
      ],
      'sticky' => [
        'type' => 'boolean',
        'title' => 'Sticky at top of lists',
        'cardinality' => 1,
      ],
      'default_langcode' => [
        'type' => 'boolean',
        'title' => 'Default translation',
        'description' => 'A flag indicating whether this is the default translation.',
        'cardinality' => 1,
      ],
      'revision_default' => [
        'type' => 'boolean',
        'title' => 'Default revision',
        'description' => 'A flag indicating whether this was a default revision when it was saved.',
        'cardinality' => 1,
      ],
      'revision_translation_affected' => [
        'type' => 'boolean',
        'title' => 'Revision translation affected',
        'description' => 'Indicates if the last edit of a translation belongs to current revision.',
        'cardinality' => 1,
      ],
      'test_date' => [
        'type' => 'string',
        'title' => 'Test date',
        'description' => 'Test date description',
        'cardinality' => 1,
      ],
      'test_datetime' => [
        'type' => 'string',
        'title' => 'Test datetime',
        'description' => 'Test datetime description',
        'cardinality' => 1,
      ],
      'test_decimal' => [
        'type' => 'number',
        'title' => 'Test decimal',
        'description' => 'Test decimal description',
        'cardinality' => 1,
      ],
      'test_float' => [
        'type' => 'number',
        'title' => 'Test float',
        'description' => 'Test float description',
        'cardinality' => 1,
      ],
      'test_email' => [
        'type' => 'string',
        'title' => 'Test email',
        'description' => 'Test email description',
        'cardinality' => 1,
      ],
      'test_string' => [
        'type' => 'string',
        'title' => 'Test string',
        'description' => 'Test string description',
        'cardinality' => -1,
      ],
      'test_allowed_values' => [
        'type' => 'string',
        'title' => 'Test allowed values',
        'description' => 'Test allowed values description',
        'cardinality' => 1,
        'allowed_values' => [
          'option1' => 'Option 1',
          'option2' => 'Option 2',
          'option3' => 'Option 3',
        ],
      ],
      'test_allowed_values_function' => [
        'type' => 'string',
        'title' => 'Test allowed values function',
        'description' => 'Test allowed values function description',
        'cardinality' => -1,
        'allowed_values' => [
          'value1' => 'Value 1',
          'value2' => 'Value 2',
          'value3' => 'Value 3',
        ],
      ],
      'test_uri' => [
        'type' => 'string',
        'title' => 'Test URI',
        'description' => 'Test URI description',
        'cardinality' => 1,
      ],
    ];
    foreach ($expected_attributes as $name => $attribute) {

      // The type depends on the attribute cardinality.
      if ($attribute['cardinality'] === 1) {
        $this->assertEquals($attribute['type'], $data['definitions']['attributes']['properties'][$name]['type']);
      }
      else {
        $this->assertEquals('array', $data['definitions']['attributes']['properties'][$name]['type']);
        $this->assertEquals($attribute['type'], $data['definitions']['attributes']['properties'][$name]['items']['type']);
      }

      // Test attribute title and description.
      $this->assertEquals($attribute['title'], $data['definitions']['attributes']['properties'][$name]['title']);
      if (isset($attribute['description'])) {
        $this->assertEquals($attribute['description'], $data['definitions']['attributes']['properties'][$name]['description']);
      }
      else {
        $this->assertArrayNotHasKey('description', $data['definitions']['attributes']['properties'][$name]);
      }

      // Test allowed values.
      $property = $data['definitions']['attributes']['properties'][$name];
      $property = $attribute['cardinality'] === 1 ? $property : $property['items'];
      $composition = $attribute['cardinality'] === 1 ? 'oneOf' : 'anyOf';
      if (isset($attribute['allowed_values'])) {
        foreach (array_keys($attribute['allowed_values']) as $i => $const) {
          $this->assertEquals($const, $property[$composition][$i]['const']);
          $this->assertEquals($attribute['allowed_values'][$const], $property[$composition][$i]['title']);
        }
      }
      else {
        $this->assertArrayNotHasKey('oneOf', $property);
        $this->assertArrayNotHasKey('anyOf', $property);
      }
    }

    // Test that required attributes are declared.
    $this->assertEquals(['title'], $data['definitions']['attributes']['required']);

    // Test that all the expected relationships are present, and have the
    // proper type, titles, and description.
    $expected_relationships = [
      'node_type' => [
        'title' => 'Content type',
        'id_type' => 'string',
        'id_title' => 'Content type ID',
      ],
      'revision_uid' => [
        'title' => 'Revision user',
        'description' => 'The user ID of the author of the current revision.',
        'id_type' => 'integer',
        'id_title' => 'User ID',
      ],
      'uid' => [
        'title' => 'Authored by',
        'description' => 'The username of the content author.',
        'id_type' => 'integer',
        'id_title' => 'User ID',
      ],
      'test_entity_reference' => [
        'title' => 'Test entity reference',
        'description' => 'Test entity reference description',
        'id_type' => 'integer',
        'id_title' => 'Content ID',
      ],
      'test_entity_reference2' => [
        'title' => 'Test entity reference 2',
        'description' => 'Test entity reference 2 description',
        'id_type' => 'integer',
        'id_title' => 'Content ID',
      ],
    ];
    foreach ($expected_relationships as $rel => $rel_info) {

      // The schema structure depends on the relationship cardinality, so we
      // vary the item that we validate accordingly.
      if ($this->expectedResources['Article content items']['relationships'][$rel]['cardinality'] === 1) {
        $item = $data['definitions']['relationships']['properties'][$rel]['properties']['data']['properties'];
      }
      else {
        $this->assertEquals('array', $data['definitions']['relationships']['properties'][$rel]['properties']['data']['type']);
        $this->assertNotEmpty($data['definitions']['relationships']['properties'][$rel]['properties']['data']['items']['properties']);
        $item = $data['definitions']['relationships']['properties'][$rel]['properties']['data']['items']['properties'];
      }

      // Test $item.meta.*.
      $this->assertEquals($rel_info['id_type'], $item['meta']['type']);
      $this->assertEquals($rel_info['id_title'], $item['meta']['title']);

      // Test relationships.properties.$name.title.
      $this->assertEquals($rel_info['title'], $data['definitions']['relationships']['properties'][$rel]['title']);

      // Test relationships.properties.$name.description.
      if (!empty($rel_info['description'])) {
        $this->assertEquals($rel_info['description'], $data['definitions']['relationships']['properties'][$rel]['description']);
      }
      else {
        $this->assertArrayNotHasKey('description', $data['definitions']['relationships']['properties'][$rel]);
      }
    }

    // Test that required relationships are declared.
    $this->assertEquals(['test_entity_reference'], $data['definitions']['relationships']['required']);
  }

  /**
   * Test specific field type schema details not covered by other tests.
   */
  public function testJsonApiSchemaFieldTypes() {

    // Request resource schema for node--article.
    $data = $this->apiSchemaRequest('/jsonapi/node/article/resource/schema');

    // Test boolean fields.
    $this->assertEquals(TRUE, $data['definitions']['attributes']['properties']['status']['default']);
    $this->assertEquals(FALSE, $data['definitions']['attributes']['properties']['sticky']['default']);

    // Test datetime fields.
    $this->assertEquals('date', $data['definitions']['attributes']['properties']['test_date']['format']);
    $this->assertEquals('date-time', $data['definitions']['attributes']['properties']['test_datetime']['format']);

    // Test email fields.
    $this->assertEquals('email', $data['definitions']['attributes']['properties']['test_email']['format']);

    // Test entity_reference fields.
    $this->assertEquals(['node--article', 'node--blog'], $data['definitions']['relationships']['properties']['test_entity_reference']['properties']['data']['items']['properties']['type']['enum']);
    $this->assertEquals(['node--article', 'node--blog', 'node--page'], $data['definitions']['relationships']['properties']['test_entity_reference2']['properties']['data']['properties']['type']['enum']);

    // Test string fields.
    $this->assertEquals(255, $data['definitions']['attributes']['properties']['title']['maxLength']);
    $this->assertEquals('', $data['definitions']['attributes']['properties']['revision_log']['default']);
    $this->assertEquals('Text value', $data['definitions']['attributes']['properties']['test_string']['items']['title']);
    $this->assertEquals(255, $data['definitions']['attributes']['properties']['test_string']['items']['maxLength']);

    // Test timestamp fields.
    $this->assertEquals('date-time', $data['definitions']['attributes']['properties']['created']['format']);

    // Test URI fields.
    $this->assertEquals('uri', $data['definitions']['attributes']['properties']['test_uri']['format']);
  }

  /**
   * Test that related schemas are generated.
   */
  public function testJsonApiSchemaRelated() {
    foreach ($this->expectedResources as $resource_info) {
      if (!empty($resource_info['relationships'])) {
        foreach ($resource_info['relationships'] as $rel => $expected) {
          $data = $this->apiSchemaRequest('/jsonapi/' . $resource_info['entity_type'] . '/' . $resource_info['bundle'] . '/resource/relationships/' . $rel . '/related/schema');
          $this->assertEquals($expected['title'], $data['title']);
          $this->assertEquals('https://jsonapi.org/schema', $data['allOf'][0]['$ref']);
          $this->assertEquals('https://jsonapi.org/schema#/definitions/success', $data['allOf'][1]['if']['$ref']);
          $this->assertEquals('object', $data['allOf'][1]['then']['type']);
          $this->assertEquals('#/definitions/data', $data['allOf'][1]['then']['properties']['data']['$ref']);
          $this->assertEquals(['data'], $data['allOf'][1]['then']['required']);
          if (is_array($expected['bundle'])) {
            foreach ($expected['bundle'] as $i => $expected_bundle) {
              if ($expected['cardinality'] === 1) {
                $this->assertEquals('http://localhost/jsonapi/' . $expected['entity_type'] . '/' . $expected_bundle . '/resource/schema', $data['definitions']['data']['anyOf'][$i]['$ref']);
              }
              else {
                $this->assertEquals('http://localhost/jsonapi/' . $expected['entity_type'] . '/' . $expected_bundle . '/resource/schema', $data['definitions']['data']['items']['anyOf'][$i]['$ref']);
              }
            }
          }
          else {
            if ($expected['cardinality'] === 1) {
              $this->assertEquals('http://localhost/jsonapi/' . $expected['entity_type'] . '/' . $expected['bundle'] . '/resource/schema', $data['definitions']['data']['$ref']);
            }
            else {
              $this->assertEquals('http://localhost/jsonapi/' . $expected['entity_type'] . '/' . $expected['bundle'] . '/resource/schema', $data['definitions']['data']['items']['$ref']);
            }
          }
        }
      }
    }
  }

  /**
   * Helper function for performing an JSON:API Schema request.
   *
   * @param string $endpoint
   *   The API endpoint.
   *
   * @return array
   *   An array of JSON-decoded data returned by the request.
   */
  protected function apiSchemaRequest(string $endpoint) {
    $http_kernel = $this->container->get('http_kernel');
    $request = Request::create($endpoint);
    $request->headers->set('Accept', 'application/vnd.api+json');
    $request->headers->set('Content-Type', 'application/vnd.api+json');
    $response = $http_kernel->handle($request);
    $this->assertEquals(Response::HTTP_OK, $response->getStatusCode());
    $data = Json::decode($response->getContent());
    $this->assertEquals('https://json-schema.org/draft/2019-09/hyper-schema', $data['$schema']);
    $this->assertEquals('http://localhost' . $endpoint, $data['$id']);
    return $data;
  }

}
